//-< DBSCLS.H >------------------------------------------------------*--------*
// GOODS                     Version 1.0         (c) 1997  GARRET    *     ?  *
// (Generic Object Oriented Database System)                         *   /\|  *
//                                                                   *  /  \  *
//                          Created:     28-Mar-97    K.A. Knizhnik  * / [] \ *
//                          Last update: 17-Apr-97    K.A. Knizhnik  * GARRET *
//-------------------------------------------------------------------*--------*
// Some common database classes
//-------------------------------------------------------------------*--------*

#ifndef __DBSCLS_H__
#define __DBSCLS_H__

#include "goods.h"

//
// Dynamicly reallocatable array template. Array is automatically
// expanded when new element is stored at position beyond array bound. 
// This template can't be used directly because it is impossible 
// to register template classes in database.
// Instead of this concrete instantiations of this template can be used. 
//
template<class T>
class array_template : public object { 
  protected: 
    nat4 n_elems;
    T    buf[1];

    virtual array_template* clone(size_t new_size) { 
	return new (self_class, new_size) array_template(0, cls, new_size); 
    }

    array_template* set_size(nat4 n) { 
	nat4 allocated = (size - offsetof(array_template, buf)) / sizeof(T); 
	if (n > allocated) { 
	    array_template* new_arr = clone(n < allocated*2 ? allocated*2 : n);
	    memcpy(new_arr->buf, buf, n_elems*sizeof(T));
	    new_arr->n_elems = n; 
	    new_arr->mop = mop;
	    become(new_arr); 
	    return new_arr; 
	}
	n_elems = n; 
	return this; 
    }
  public:
    static ref<array_template<T> > create(size_t init_size = 0) { 
	return new (self_class,init_size) array_template(self_class,init_size);
    }

    size_t length() const { return n_elems; }

    //
    // Array elements accessors
    //
    void putat(nat4 index, T elem) { 
	if (index >= n_elems) { 
	    set_size(index+1)->buf[index] = elem; 
	} else { 
	    buf[index] = elem; 
	}
    }  
    T getat(nat4 index) const { 
	assert(index < n_elems); 
	return buf[index]; 
    }  
    T operator[](nat4 index) const {  
	assert(index < n_elems); 
	return buf[index]; 
    }  
    
    //
    // Stack protocol
    //
    void push(T elem) { 
	int n = n_elems; 
	set_size(n+1)->buf[n] = elem;
    }
    T pop() { 
	assert(n_elems > 0);
	n_elems -= 1;
	return buf[n_elems]; 
    }
    T top() const { 
	assert(n_elems > 0);
	return buf[n_elems-1];
    }
    
    //
    // Massive operation
    // 
    void insert(nat4 pos, T elem) { 
	nat4 n = n_elems;
	assert(pos <= n);
	array_template* arr = set_size(n+1);
	memmove(&arr->buf[pos+1], &arr->buf[pos], (n - pos)*sizeof(T));
	memset(&arr->buf[pos], 0, sizeof(T)); 
	arr->buf[pos] = elem; 
    }

    void append(T const* tail, size_t len) { 
	nat4 n = n_elems;
	array_template* arr = set_size(n+len);
	for (size_t i = 0; i < len; i++) { 
	    arr->buf[i+n] = tail[i];
	}
    }	

    void remove(nat4 pos) {
	assert(pos < n_elems);
	T zero;
	buf[pos] = zero;
	memcpy(&buf[pos], &buf[pos+1], n_elems - pos - 1); 
    }

    void copy(T* buf, int from, int len) const { 
	assert(unsigned(from + len) <= n_elems && from >= 0 && len >= 0);
	int n = len; 
	while (--n >= 0) { 
	    buf[n] = this->buf[n + from]; 
	}
    }
 
    METACLASS_DECLARATIONS(array_template, object); 

  protected: 
    array_template(class_descriptor& desc, size_t init_size) 
    : object(desc, init_size*sizeof(T)) 
    {
	n_elems = init_size;
    }
};

template<class T>
inline field_descriptor& array_template<T>::describe_components() { 
    return FIELD(n_elems), VARYING(buf); 
}

typedef array_template<char>          arrayOfChar; 
typedef array_template<int4>          arrayOfInt; 
typedef array_template<real8>         arrayOfReal; 
typedef array_template< ref<object> > arrayOfObject; 

class string : public arrayOfChar { 
  protected: 
    virtual array_template<char>* clone(size_t new_size) { 
	return new (self_class, new_size) string(0, cls, new_size); 
    }

  public: 
    int  compare(const char* str) const { return strcmp(buf, str); }
    int  compare(ref<string> const& str) const { return -str->compare(buf); }

    int  index(const char* str) const { 
	char* p = strstr(buf, str); 
	return p ? p - buf : -1; 
    }
    
    void append(const char* tail) { 
	nat4 n = n_elems ? n_elems-1 : 0;
	size_t len = strlen(tail) + 1;
	string* s = (string*)set_size(n+len);
	for (size_t i = 0; i < len; i++) { 
	    s->buf[i+n] = tail[i];
	}
    }	
	
    static ref<string> create(size_t init_size = 0) { 
	return new (self_class, init_size) string(self_class, init_size); 
    }

    static ref<string> create(const char* str) { 
	size_t len = strlen(str)+1;
	string* s = new (self_class, len) string(self_class, len); 
	strcpy(s->buf, str);
	return s;
    }

    void replace(const char* str) { 
	string* s = (string*)set_size(strlen(str)+1);
	strcpy(s->buf, str);
    }

    const char* get_text() const {
        // it is safe to use pointer to buffer of persistent string 
        // only within transaction in which this object was included
        return buf;
    }

    void print() const { console::output(buf); }

    METACLASS_DECLARATIONS(string, arrayOfChar); 

  protected: 
    string(class_descriptor& desc, size_t init_size) 
    : arrayOfChar(desc, init_size) {}
}; 


//
// Representation of one-to-many relationship: set of objects.
// Set conssist of set owner and a number of set members. 
// Each member contains references to previouse and next member, 
// reference to set owner, reference to associated object 
// and full key of this object. First (last) member has next (prev) field 
// equal to NULL. Member of a set has virtual function to compare
// two members. Also default function for converting complete key 
// to short integer form
//

typedef nat4 skey_t;

class set_member : public object {
    friend class set_owner;
  public: 
    ref<set_member> next; 
    ref<set_member> prev; 
    ref<set_owner>  owner; 
    ref<object>     obj;

    char            key[1];

    static ref<set_member> create(ref<object> obj, const char* key) { 
	size_t key_size = strlen(key) + 1;
	return new (self_class, key_size) set_member(obj, key, key_size);
    }
    static ref<set_member> create(ref<object> obj, const char* key, 
				  size_t key_size) { 
	return new (self_class, key_size) set_member(obj, key, key_size);
    }

    ref<set_member> first() const;  
    ref<set_member> last() const;  

    inline  int     compare(ref<set_member> mbr) const { 
	return -mbr->compare(key, size - offsetof(set_member, key)); 
    }
    virtual int     compare(const char* key, size_t size) const; 
    virtual skey_t  get_key() const; 
    
    static skey_t   str2key(const char* s, size_t len);

    METACLASS_DECLARATIONS(set_member, object);

  protected: 

    set_member(ref<object> const& obj, const char* key, size_t key_size) 
    : object(self_class, key_size)
    { 
	next = NULL;
	prev = NULL; 
	owner = NULL; 
	this->obj = obj; 
	memcpy(this->key, key, key_size);
    }
}; 


class set_owner : public object { 
  public: 
    ref<set_member> first; 
    ref<set_member> last; 
    ref<object>     obj;
    nat4            n_members;
    
    bool empty() const { return first == last; }
    ref<set_member> find(const char* key) const; 

    void put_first(ref<set_member> mbr);
    void put_last(ref<set_member> mbr);
    void put_after(ref<set_member> after, ref<set_member> mbr);
    void put_before(ref<set_member> before, ref<set_member> mbr);

    typedef void (*member_function)(ref<set_member> mbr);
    void iterate(member_function) const;

    ref<set_member> get_first(); 
    ref<set_member> get_last(); 
    void remove(ref<set_member> mbr);

    static ref<set_owner> create(ref<object> obj) { 
	return new set_owner(self_class, obj); 
    } 

    METACLASS_DECLARATIONS(set_owner, object);

  protected: 
    set_owner(class_descriptor& desc, ref<object> const& obj, int varying=0) 
    : object(desc, varying)
    { 
	this->obj = obj; 
	n_members = 0;
    }
};  

//
// B tree - multibranch balanced tree, provides fast access for objects 
// in database. B_tree class is derived from set_owner and use set members
// to reference objects. 
//
class B_tree : public set_owner {
  public:
    ref<set_member> find(skey_t key) const;

    ref<set_member> find(const char* str, size_t len, skey_t key) const;
    ref<set_member> find(const char* str) const;

    void            insert(ref<set_member> mbr);
    void            remove(ref<set_member> mbr);

    void            dump() const;
    void            ok() const;

    static ref<B_tree> create(ref<object> obj) { 
	return new B_tree(self_class, obj); 
    } 

    METACLASS_DECLARATIONS(B_tree, set_owner);

  protected:
    friend class B_page;

    ref<B_page> root;   // root page
    nat4        height; // height of tree

    B_tree(class_descriptor& desc, ref<object> const& obj=NULL, int varying=0) 
    : set_owner(desc, obj, varying) 
    {
	root = NULL; 
	height = 0;
    }
}; 

class B_page : public object { 
    friend class B_tree;
  protected: 
    // enum { n = 408 }; // Minimal number of used items at not root page 
    enum { n = 2 };    // only for testing
    int4 m;            // Index of first used item

    struct item { 
	ref<object> p; // for leaf page - pointer to 'set_member' object
	               // for all other pages - pointer to child page
	skey_t      key; 

	field_descriptor& describe_components() { 
	    return FIELD(p), FIELD(key);
	}				
	inline friend field_descriptor& describe_field(item& s) { 
	    return s.describe_components();
	}
    } e[n*2];
    //
    // Class invariants: 
    //
    //   0 <= m <= 2*n-2, for root page
    //   0 <= m <= n, for other pages
    //
    // For not leaf page:
    // ( for all i : m < i < 2*n : e[i-1].key < e[i].key ) &&
    // ( for all i, j : m <= i < 2*n, e[i].p->m <= j < 2*n-1 :
    //   e[i].key > e[i].p->e[j].key &&
    //   e[i].key == e[i].p->e[2*n-1].key )
    //

    ref<set_member> find(int height, skey_t key) const; 

    enum operation_effect { 
	done, 
	overflow,
	underflow, 
	propagation
    };  

    operation_effect insert(int r, item& ins);
    operation_effect remove(int r, item& rem);
    operation_effect handle_underflow(int r, skey_t key, item& rem);

    operation_effect find_insert_position(int level, 
					  ref<B_tree> const& tree, 
					  item& ins) const;
    operation_effect find_remove_position(int level, item& rem) const;

    void dump(int level, int height) const;
    void ok(skey_t min_key_val, skey_t max_key_val,
	    int level, int height) const;

    B_page() : object(self_class) {}
    B_page(ref<B_page> const& root, item& ins); // create new B tree root

  public: 
    METACLASS_DECLARATIONS(B_page, object);
}; 

//
// Simple hash table.
//
class hash_item : public object { 
  public: 
    ref<hash_item> next;
    ref<object>    obj;
    char           name[1];

    //
    // Call function for object. Apply method returns value
    // returned by function. Iteration through all hash table 
    // elements will finished if true value is return by apply method.
    //
    typedef bool (*item_function)(const char* name, ref<object> obj);
    bool apply(item_function f) const { return f(name, obj); }

    hash_item(ref<object> obj, ref<hash_item> next, 
	      const char* name, size_t name_len) 
    : object(self_class, name_len)
    {
	this->obj = obj;
	this->next = next;
	strcpy(this->name, name);
    }
    int compare(const char* name) const { 
	return strcmp(name, this->name);
    }
    static ref<hash_item> create(ref<object> obj, ref<hash_item> chain,
				 const char* name)
    {
	size_t name_len = strlen(name)+1;
	return new (self_class,name_len) hash_item(obj, chain, name, name_len);
    }
    METACLASS_DECLARATIONS(hash_item, object);
};

class hash_table : public object { 
  public: 
    //
    // Add to hash table association of object with specified name.
    //
    void        put(const char* name, ref<object> obj);
    //
    // Search for object with specified name in hash table.
    // If object is not found NULL is returned.
    //
    ref<object> get(const char* name) const;
    //
    // Remove object with specified name from hash table.
    // If there are several objects with the same name the one last inserted
    // is removed. If such name was not found 'false' is returned.
    //
    bool        del(const char* name);
    //
    // Remove concrete object with specified name from hash table.
    // If such name was not found or it is associated with different object
    // 'false' is returned. If there are several objects with the same name, 
    // all of them are compared with specified object.
    //    
    bool        del(const char* name, ref<object> obj);

    //
    // Apply function to objects placed in hash table.
    // Function will be applied in some unspecified order
    // to all elements in hash table or until true will be returned by 
    // the function. In the last case reference to this object is returned.
    // Otherwise (if function returns false for all elements in hash table)
    // NULL is returned.
    //
    ref<object> apply(hash_item::item_function) const;

    void reset() { 
	n_elems = 0;
	memset(table, 0, size*sizeof(ref<object>)); 
    }

    size_t get_numer_of_elements() const { return n_elems; }

    static ref<hash_table> create(size_t size) {
	return new (self_class, size) hash_table(size);
    }
    METACLASS_DECLARATIONS(hash_table, object);

  protected: 
    nat4        size;
    nat4        n_elems;
    ref<object> table[1];
    
    hash_table(size_t size, class_descriptor& desc = self_class) 
    : object(desc, size) 
    { 
	this->size = size;
	n_elems = 0;
    }
};

//
// Hash tree is combination of hash table and B-tree. Calculated 
// values of hash function are placed at B*-tree pages.
//
class H_page : public object { // page of hash tree
  public:
    enum { size_log=5 };
    ref<object> p[1 << size_log];

    ref<object> apply(hash_item::item_function f, int height) const;
    void reset(int height);
    H_page() : object(self_class) {}

    METACLASS_DECLARATIONS(H_page, object);
};

class H_tree : public object { 
  public: 
    //
    // Add to hash table association of object with specified name.
    //
    void        put(const char* name, ref<object> obj);
    //
    // Search for object with specified name in hash table.
    // If object is not found NULL is returned.
    //
    ref<object> get(const char* name) const;
    //
    // Remove object with specified name from hash table.
    // If there are several objects with the same name the one last inserted
    // is removed. If such name was not found 'false' is returned.
    //
    bool        del(const char* name);
    //
    // Remove concrete object with specified name from hash table.
    // If such name was not found or it is associated with different object
    // 'false' is returned. If there are several objects with the same name, 
    // all of them are compared with specified object.
    //    
    bool        del(const char* name, ref<object> obj);

    //
    // Apply function to objects placed in hash table.
    // Function will be applied in some unspecified order
    // to all elements in hash table or until true will be returned by 
    // the function. In the last case reference to this object is returned.
    // Otherwise (if function returns false for all elements in hash table)
    // NULL is returned.
    //
    ref<object> apply(hash_item::item_function f) const { 
	if (!root.is_nil()) {
	    return root->apply(f, height);
	}
	return NULL;
    }
    //
    // Clear hash table
    //
    void reset() { 
	if (!root.is_nil()) { 
	    modify(root)->reset(height);
	}
	n_elems = 0;
    }
    size_t get_numer_of_elements() const { return n_elems; }

    METACLASS_DECLARATIONS(H_tree, object);
    H_tree(size_t hash_size, class_descriptor& desc = self_class);

  protected: 
    nat4        size;
    nat4        height;
    nat4        n_elems;
    ref<H_page> root;
};

//
// Binary large object class. This class can be used as base for
// creating multimedia or text objects with large size and sequential 
// access. This class provide functionality for incremental object
// loading by parallel task. 
//

class blob : public object { 
  protected: 
    ref<blob> next;
    ref<blob> last;
    nat4      size;
    nat1      data[1];
    
    static void task_proc load_next_blob_part(void* arg);

  public: 
    //
    // This method is called by 'play' method for each part of blob
    // object within conext of current task. If this method returns false 
    // preloading process is terminated. 
    //
    virtual bool handle() const;
    //
    // Default implementation of this method just extracts all parts of 
    // blob from database and calls handle method for each part.
    //
    virtual void play() const;

    //
    // Append new blob object at the end of blob objects chain.
    //
    void append(ref<blob> bp) { 
	modify(last)->next = bp; 
	last = bp;
    }

    METACLASS_DECLARATIONS(blob, object);

    blob(size_t part_size, class_descriptor& desc = self_class) 
    : object(desc, part_size) 
    { 
	next = NULL;
	last = this;
	size = part_size;
    }

    blob(void* buf, size_t buf_size, class_descriptor& desc = self_class) 
    : object(desc, buf_size) 
    { 
	next = NULL;
	last = this;
	size = buf_size;
	memcpy(data, buf, buf_size);
    }
};

#endif
